import time
import json
import logging
import asyncio
import aiobotocore
from .worker import BatchWorker
logger = logging.getLogger(__name__)


class Firehose:
    service_name = 'firehose'
    MAX_RETRY = 10

    def __init__(self, stream_name, region_name,
                 encode_func=None, delimiter=b'\n',
                 *, worker_params=None, client_params=None,
                 original_api=False):
        self.session = aiobotocore.get_session()
        self.stream_name = stream_name
        self.encode_func = encode_func or (lambda o: json.dumps(o).encode())
        self.delimiter = delimiter
        client_params = client_params or {}
        client_params['region_name'] = region_name
        self.client = self.session.create_client(
            self.service_name, **client_params)
        worker_params = worker_params or {}
        batch_func = self.original_batch if original_api else self.write_batch
        self.worker = BatchWorker(batch_func, **worker_params)

    def __repr__(self):
        return '<{}: stream={}, worker={!r}>'.format(
                self.__class__.__name__, self.stream_name, self.worker)

    async def put(self, obj):
        return await self.worker.put(obj)

    async def stop(self):
        timestamp = time.time()
        await self.worker.stop()
        await self.client.close()
        cost = time.time() - timestamp
        logger.info('{0!r} stopped in {1:.1f} seconds'.format(self, cost))

    def _encode(self, obj_list):
        encoded = [self.encode_func(obj) for obj in obj_list]
        encoded.append(b'')
        return self.delimiter.join(encoded)

    async def write_batch(self, obj_list):
        try:
            data = self._encode(obj_list)
            await self.client.put_record(
                DeliveryStreamName=self.stream_name,
                Record={'Data': data}
            )
        except Exception as e:
            logger.exception(e)
            return obj_list

    async def original_batch(self, obj_list, _seq=0):
        if _seq > 0:
            await asyncio.sleep(0.1 * (2 ** _seq))
        if _seq > self.MAX_RETRY:
            raise Exception(
                'write_batch error: firehose put_record_batch failed')
        try:
            resp = await self.client.put_record_batch(
                DeliveryStreamName=self.stream_name,
                Records=[{
                    'Data': self.encode_func(obj) + self.delimiter
                } for obj in obj_list]
            )
        except Exception as e:
            logger.exception(e)
            return obj_list
        if resp['FailedPutCount'] > 0:
            failed_obj_list = [
                obj_list[i] for i, record in
                enumerate(resp['RequestResponses'])
                if 'ErrorCode' in record]
            return await self.write_batch(
                failed_obj_list, _seq=_seq + 1
            )



import time
import logging
import asyncio
import async_timeout
logger = logging.getLogger(__name__)


class BatchWorker:
    def __init__(self, handler, *, maxsize=0, aggr_num=10, timeout=5,
                 concurrency=10):
        self.queue = asyncio.Queue(maxsize)
        self.handler = handler
        self.aggr_num = aggr_num
        self.timeout = timeout
        self.concurrency = concurrency
        self.semaphore = asyncio.Semaphore(concurrency)
        self.quit = object()
        self.shutdown = False
        self.fut = None
        self.futures = set()
        self.start()

    def __repr__(self):
        return '<{}: qsize={}, concurrency={}, working={}>'.format(
            self.__class__.__name__, self.queue.qsize(), self.concurrency,
            len(self.futures)
        )

    @property
    def qsize(self):
        return self.queue.qsize()

    async def put(self, obj):
        await self.queue.put(obj)

    def start(self):
        self.fut = asyncio.ensure_future(self.run())

    async def stop(self):
        logger.info('Stopping {0!r}'.format(self))
        await self.queue.put(self.quit)
        if self.fut:
            await self.fut

    async def _get_obj_list(self):
        obj_list = []
        timeout = self.timeout
        while timeout > 0 and len(obj_list) < self.aggr_num:
            timestamp = time.time()
            try:
                async with async_timeout.timeout(timeout):
                    obj = await self.queue.get()
            except asyncio.TimeoutError:
                break
            if obj is self.quit:
                self.shutdown = True
                break
            obj_list.append(obj)
            timeout -= time.time() - timestamp
        return obj_list

    async def run(self):
        logger.info('Starting {0!r}'.format(self))
        while not self.shutdown:
            obj_list = await self._get_obj_list()
            if not obj_list:
                continue
            await self.semaphore.acquire()
            fut = asyncio.ensure_future(self.handle(obj_list))
            self.futures.add(fut)
            fut.add_done_callback(self.futures.remove)
        if self.futures:
            await asyncio.wait(self.futures)

    async def handle(self, obj_list):
        try:
            try:
                result = self.handler(obj_list)
                if asyncio.iscoroutine(result):
                    result = await result
                if result:
                    for obj in result:
                        await self.queue.put(obj)
            except Exception as e:
                logger.exception(e)
        finally:
            self.semaphore.release()



import time
import logging
from .worker import BatchWorker

logger = logging.getLogger(__name__)


class DyanmodbWriter:
    def __init__(self, table, *, worker_params=None):
        self.table = table
        worker_params = worker_params or {}
        self.worker = BatchWorker(self.write_batch, **worker_params)

    def __repr__(self):
        return "<{}: table={}, worker={!r}>".format(
            self.__class__.__name__, self.table, self.worker
        )

    async def stop(self):
        timestamp = time.time()
        await self.worker.stop()
        cost = time.time() - timestamp
        logger.info("{0!r} stopped in {1:.1f} seconds".format(self, cost))

    async def write_one(self, Item):
        await self.worker.put(Item)

    async def write_batch(self, obj_list):
        async with self.table.batch_writer() as batch:
            for Item in obj_list:
                await batch.put_item(Item=Item)
